﻿//
//  machine.cpp
//  Yon
//
//  Created by Miso on 13.8.2009.
//  Copyright 2009 Miso Antonic. All rights reserved.
//
//

#include "machine.h"
#include "language.h"
//#include "assert.h"
#include <iostream>
#include "math.h"

#include "module.h"



#define TAIL_CALL_OPTIMIZATION_ON false
//#define BENCHMARK

bool haltSignal = false;

ULONG symbolForCString(CHAR** symbolTable, ULONG symbolTableLength, const CHAR* s)
{
	assert(s != NULL);
	ULONG symbol;
	for (symbol=0; symbol<symbolTableLength; symbol++)
		if (wcscmp(symbolTable[symbol], s) == 0)
			return symbol;
	
	return 0;
}

void printPool(ObjectRef* pool)
{
	//ObjectRef o = machine->pool;
	wprintf(L"--- POOLS:\n");
    int i;
    for (i=0; i<numberOfObjectTypes; i++)
    {
        wprintf(L"  POOL %d:\n", i);
        ObjectRef o = pool[i];
        int count = 0;
        while (o)
        {
            wprintf(L"    (%d next: %d)\n", o, o->next);
            o = o->next;
            count++;
        }
        wprintf(L"    POOL count: %d\n", count);
    }
}


const CHAR* objectTypeToCString(BYTE type)
{
	if (type == typeReal)
		return L"REAL";
	if (type == typeCall)
		return L"CALL";
	if (type == typeBigCall)
		return L"BIG_CALL";
	if (type == typeNil)
		return L"NIL";
	if (type == typeFunction)
		return L"FUNCTION";
	if (type == typeBool)
		return L"BOOLEAN";
	if (type == _typeProperty)
		return L"PROPERTY";
	if (type == typeNativeFunction)
		return L"NATIVE_FUNCTION";
	if (type == typeObject)
		return L"OBJECT";
	if (type == typeString)
		return L"STRING";
	if (type == typeArray)
		return L"ARRAY";
	if (type == _typeData)
		return L"DATA";
	if (type == typeInternalPrototype)
		return L"INTERNAL_PROTOTYPE";
	return L"!UNKNOWN!";
}



void printObject(MachineRef machine, ObjectRef object)
{
	if (!object)
	{
		wprintf(L"NULL\n");
		return;
	}
	wprintf(L"%ls (%#x) ", objectTypeToCString(object->type), object);
    if (object->tag & TAG_GARBAGE)
        wprintf(L"GARBAGE!!!");
	ULONG i;
	if (object->parent == machine->objectPrototype)
		wprintf(L"!!! PROTOTYPE !!!");
	wprintf(L"\n");
	if (object->type == typeBool)
	{
		wprintf(L"   bool value: %ls\n", BOOL(object)->value ? L"YES" : L"NO");
	}
	if (object->type == typeReal)
	{
		wprintf(L"   real value: %f\n", REAL(object)->value);
	}
	if (object->type == typeFunction)
	{
		wprintf(L"   start bytecode index: %d\n", FUNCTION(object)->startAddress - machine->bytecodes);
        
		wprintf(L"   number of parameters: %d\n", FUNCTION(object)->numberOfParameters);
		if (FUNCTION(object)->parentCall)
			wprintf(L"   parent call: %ls (%#x)\n", objectTypeToCString(FUNCTION(object)->parentCall->type), FUNCTION(object)->parentCall);
		else
			wprintf(L"   parent call: NULL\n");
		if (FUNCTION(object)->frozenCall)
			wprintf(L"   frozen call: %ls (%#x)\n", objectTypeToCString(FUNCTION(object)->frozenCall->type), FUNCTION(object)->frozenCall);
		else
			wprintf(L"   frozen call: NULL\n");
        
		wprintf(L"   parameters: ");
		for (i=0; i<FUNCTION(object)->numberOfParameters; i++)
			wprintf(L"%ls ", machine->symbolTable[FUNCTION(object)->parameters[i]]);
	    wprintf(L"\n");
	}
	if (object->type == typeNativeFunction)
	{
		wprintf(L"   address: %x\n", NATIVE_FUNCTION(object)->address);
		wprintf(L"   number of parameters: %d\n", NATIVE_FUNCTION(object)->numberOfParameters);
	}
	if (object->type == typeCall || object->type == typeBigCall)
	{
		wprintf(L"   slot capacity: %d\n", CALL(object)->slotCapacity);
		wprintf(L"   stack capacity: %d\n", CALL(object)->stackCapacity);
		wprintf(L"   call type: %d\n", CALL(object)->callType);
		
		if (CALL(object)->parentCall)
			wprintf(L"   parent call: %ls (%#x)\n", objectTypeToCString(CALL(object)->parentCall->type), CALL(object)->parentCall);
		else
			wprintf(L"   parent call: NULL\n");
		
		if (CALL(object)->bytecodeAddress)
			wprintf(L"   bytecode index: %d\n", CALL(object)->bytecodeAddress - machine->bytecodes);
		else
			wprintf(L"   bytecode index: NULL\n");
		
		if (CALL(object)->calledFunction)
			wprintf(L"   called function: %ls (%#x)\n", objectTypeToCString(CALL(object)->calledFunction->type), CALL(object)->calledFunction);
		else
			wprintf(L"   called function: NULL\n");
        
        if (CALL(object)->yieldedFromCall)
			wprintf(L"   yieldedFromCall: %ls (%#x)\n", objectTypeToCString(CALL(object)->yieldedFromCall->type), CALL(object)->yieldedFromCall);
		else
			wprintf(L"   yieldedFromCall: NULL\n");
		
		if (CALL(object)->self)
			wprintf(L"   self: %ls (%#x)\n", objectTypeToCString(CALL(object)->self->type), CALL(object)->self);
		else
			wprintf(L"   self: NULL\n");

		if (CALL(object)->calledFunction)
			wprintf(L"   self: %ls (%#x)\n", objectTypeToCString(CALL(object)->calledFunction->type), CALL(object)->calledFunction);
		else
			wprintf(L"   self: NULL\n");
        
		wprintf(L"   stack:\n");
		if (CALL(object)->stackTop)
		{
			ObjectRef* stackItem;
			for (stackItem = CALL(object)->stackTop; stackItem >= CALL(object)->stack; stackItem--)
				wprintf(L"      %ls (%#x)\n", objectTypeToCString((*stackItem)->type), *stackItem);
		}
		else
			wprintf(L"       ACTIVE\n");
	}
	if (object->type == typeString)
	{
		wprintf(L"   length: %d\n", STRING(object)->size);
		wprintf(L"   value: %ls\n", STRING(object)->data);
	}
	if (object->type == typeArray)
	{
		wprintf(L"   length: %d\n", ARRAY(object)->size);
		wprintf(L"   items:\n");
		for (i=0; i<ARRAY(object)->size; i++)
			wprintf(L"      %d: %ls (%#x)\n", i, objectTypeToCString(ARRAY(object)->data[i]->type), ARRAY(object)->data[i]);
	}
    
    if (object->type == _typeData)
	{
		wprintf(L"   size: %d\n", DATA(object)->size);
	}
    //	wprintf(L"   prev: %#x\n", object->prev);
    //	wprintf(L"   next: %#x\n", object->next);
	if (object->slotCount)
	{
		wprintf(L"   slots:\n");
		for (i=0; i<object->slotCount; i++)
		{
			wprintf(L"   Symbol: %d\n", object->slots[i].symbol);
			wprintf(L"      %ls: %ls (%#x)\n", machine->symbolTable[object->slots[i].symbol], objectTypeToCString(object->slots[i].value->type), object->slots[i].value);
		}
	}
}

void printObjects(MachineRef machine)
{
	ObjectRef o = machine->objects;
	wprintf(L"\nOBJECTS:\n");
	int count = 0;
	while (o)
	{
		printObject(machine, o);
        //		wprintf(L"(%ls %d prev: %d next: %d)\n", objectTypeToCString(o->type), o, o->prev, o->next);
		o = o->next;
		count++;
	}
	wprintf(L"OBJECTS count: %d\n", count);
}


ObjectRef newObjectFromPool(ObjectRef* objects, ObjectRef* pool, BYTE type, ObjectRef parent, ULONG size)
{
    ObjectRef object;
	if (!pool[type])
    {
        // No such object in pool, let's make a new one
        object = (ObjectRef)malloc(size);
        assert(object != NULL);
        object->slots = NULL;
        object->slotCount = 0;
        object->tag = 0;
    }
    else
    {
        object = pool[type];
        pool[type] = object->next;        
    }
    object->tag &= ~TAG_GARBAGE; 
    
	object->type = type;
	object->parent = parent;
	object->next = *objects;
	*objects = object;
	object->prev = NULL;
	if (object->next)
		object->next->prev = object;
    
	return object;
}

static void freeCallObjectToPool(MachineRef machine, CallObjectRef object)
{
    
	object->slotCount = 0;
	object->tag |= TAG_GARBAGE;
	
//	if (object->prev)
//		object->prev->next = object->next;
//	else
//		machine->objects = object->next;
//	if (object->next)
//		object->next->prev = object->prev;
	
	object->next = (ObjectRef)machine->callPools[object->slotCapacity*OBJECTREFS_IN_SLOT + object->stackCapacity];
	machine->callPools[object->slotCapacity*OBJECTREFS_IN_SLOT + object->stackCapacity] = (CallObjectRef)object;
}

static void freeObjectToPool(MachineRef machine, ObjectRef object)
{	
	if (object->type == typeBigCall)
		free(((CallObjectRef)object)->stack);
    else if (object->type == typeString)
    {
        free(((StringObjectRef)object)->data);
    }
    else if (object->type == typeArray)
    {
        free(((ArrayObjectRef)object)->data);
    }
    else if (object->type == _typeData)
    {
        BYTE* bytes = ((DataObjectRef)object)->bytes;
        free(bytes);
    }
	
	if (object->slots)
	{
		free(object->slots);
	}
	
    object->tag |= TAG_GARBAGE;
	object->slots = NULL;
	object->slotCount = 0;
	object->tag = 0;
	
//	if (object->prev)
//		object->prev->next = object->next;
//	else
//		machine->objects = object->next;
//	if (object->next)
//		object->next->prev = object->prev;
	
	object->next = machine->pool[object->type];
	machine->pool[object->type] = object;
}

RealObjectRef newRealObject(ObjectRef* objects, ObjectRef* pool, DOUBLE value, ObjectRef parent)
{
	RealObjectRef object = (RealObjectRef)newObjectFromPool(objects, pool, typeReal, parent, sizeof(RealObject));
	object->value = value;
	return object;
}

ClassObjectRef newClassObject(ObjectRef* objects, ObjectRef* pool, ObjectRef instancePrototype, ObjectRef parent)
{
	ClassObjectRef object = (ClassObjectRef)newObjectFromPool(objects, pool, typeClass, parent, sizeof(ClassObject));
	object->instancePrototype = instancePrototype;
	return object;
}

PropertyObjectRef newPropertyObject(ObjectRef* objects, ObjectRef* pool, ULONG readSymbol, ULONG writeSymbol, ObjectRef parent)
{
	PropertyObjectRef object = (PropertyObjectRef)newObjectFromPool(objects, pool, _typeProperty, parent, sizeof(PropertyObject));
	object->readSymbol = readSymbol;
	object->writeSymbol = writeSymbol;
	return object;
}

NativeFunctionObjectRef newNativeFunction(MachineRef machine, ObjectRef* objects, ObjectRef* pool, ULONG symbol, ObjectRef parent)
{
	NativeFunctionObjectRef object = (NativeFunctionObjectRef)newObjectFromPool(objects, pool, typeNativeFunction, parent, sizeof(NativeFunctionObject));
	if (getNativeFunction(machine->symbolTable[symbol], &object->address, &object->numberOfParameters))
		return object;
	else
		return NULL;
}

BoolObjectRef newBoolObject(ObjectRef* objects, ObjectRef* pool, bool value, ObjectRef parent)
{
	BoolObjectRef object = (BoolObjectRef)newObjectFromPool(objects, pool, typeBool, parent, sizeof(BoolObject));
	object->value = value;
	return object;
}

DataObjectRef newDataObject(ObjectRef* objects, ObjectRef* pool, BYTE* bytes, ULONG size, ObjectRef parent)
{
    DataObjectRef object = (DataObjectRef)newObjectFromPool(objects, pool, _typeData, parent, sizeof(DataObject));
    object->bytes = bytes;
    object->size = size;

    return object;
}

StringObjectRef newStringObject(ObjectRef* objects, ObjectRef* pool, const CHAR* data, ULONG length, ObjectRef parent)
{
	StringObjectRef object = (StringObjectRef)newObjectFromPool(objects, pool, typeString, parent, sizeof(StringObject));

	object->size = length;
    object->data = NULL;

	NEW_ARRAY(object->data, CHAR, length+1);
	if (data)
		wcsncpy(object->data, data, length);
    object->data[length] = NULL;
	return object;
}

StringObjectRef newStringObjectDontCopy(ObjectRef* objects, ObjectRef* pool, CHAR* data, ULONG length, ObjectRef parent)
{
	StringObjectRef object = (StringObjectRef)newObjectFromPool(objects, pool, typeString, parent, sizeof(StringObject));
	object->size = length;
    object->data = data;
	return object;
}

StringObjectRef newStringObject(MachineRef machine, const CHAR* data, ULONG length)
{
    return newStringObject(&machine->objects, machine->pool, data, length, machine->objectStringPrototype);
}

ArrayObjectRef newArrayObject(MachineRef machine, ObjectRef* data, ULONG length)
{
    return newArrayObject(&machine->objects, machine->pool, data, length, machine->objectArrayPrototype);
}

ArrayObjectRef newArrayObject(ObjectRef* objects, ObjectRef* pool, ObjectRef* data, ULONG length, ObjectRef parent)
{
	ArrayObjectRef object = (ArrayObjectRef)newObjectFromPool(objects, pool, typeArray, parent, sizeof(ArrayObject));
	object->size = length;
	if (length)
	{
		NEW_ARRAY(object->data, ObjectRef, object->size);
		if (data)
		{
			data += length;
			while (length--)
				object->data[length] = *(--data);
		}
	}
	else
		object->data = NULL;

	return object;
}

static CallObjectRef newCallObject(ULONG numberOfSlotsNeeded, ULONG stackSizeNeeded, ObjectRef* objects, ObjectRef* pool, CallObjectRef (*callPools)[MAX_CALL_POOL_DATA_SIZE], ObjectRef parent)
{
	CallObjectRef object;
	CallObjectRef* matchingCall;
	
	BYTE type;
	if (numberOfSlotsNeeded <= MAX_CALL_POOL_SLOT_SIZE && stackSizeNeeded <= MAX_CALL_POOL_STACK_SIZE)
		type = typeCall;
	else
    {
		type = typeBigCall;
		object = (CallObjectRef)newObjectFromPool(objects, pool, type, parent, sizeof(CallObject));
		object->stack = NULL;        
		REALLOC_ARRAY(object->stack, ObjectRef, stackSizeNeeded);
		REALLOC_ARRAY(object->slots, Slot, numberOfSlotsNeeded);
		object->slotCapacity = numberOfSlotsNeeded;
		object->stackCapacity = stackSizeNeeded;
		return object;
	}
    
    if (*(matchingCall = &((*callPools)[numberOfSlotsNeeded*OBJECTREFS_IN_SLOT + stackSizeNeeded])) == NULL)
	{
		object = (CallObjectRef)newObjectFromPool(objects, pool, type, parent, sizeof(CallObject)+sizeof(Slot)*numberOfSlotsNeeded+sizeof(ObjectRef)*stackSizeNeeded);
	}
    else
    {
        object = *matchingCall;
        *matchingCall = (CallObjectRef)object->next;
        object->parent = parent;
        object->next = *objects;
        *objects = (ObjectRef)object;
        object->prev = NULL;
        if (object->next)
            object->next->prev = (ObjectRef)object;
    }
    
    if (numberOfSlotsNeeded > 0)
        object->slots = (Slot*)((BYTE*)object + sizeof(CallObject));
    else
        object->slots = NULL;
    
    if (stackSizeNeeded > 0)
        object->stack = (ObjectRef*)((BYTE*)object + sizeof(CallObject) + sizeof(Slot)*numberOfSlotsNeeded);
    else
        object->stack = NULL;
    
    object->slotCapacity = numberOfSlotsNeeded;
    object->stackCapacity = stackSizeNeeded;
    object->tag &= ~TAG_GARBAGE;
	return object;
}

#define SET_DEBUG(o)  ((o)->tag |= TAG_DEBUG)
#define UNSET_DEBUG(o)  ((o)->tag &= ~TAG_DEBUG)
#define DEBUGGED(o)   ((o)->tag & TAG_DEBUG)

#define VISIT(o)	  ((o)->tag |= TAG_VISITED)
#define CACHE(o)	  ((o)->tag |= TAG_CACHED)
#define UNVISIT(o)	  ((o)->tag &= ~TAG_VISITED)
#define VISITED(o)    ((o)->tag & TAG_VISITED)
#define CACHED(o)	  ((o)->tag & TAG_CACHED)
#define DESCHEDULE(o) \
{ \
    o = scheduledObjects; \
    scheduledObjects = scheduledObjects->next; \
    \
    o->prev = NULL; \
    o->next = reachedObjects; \
    reachedObjects = o; \
    if (o->next) \
        o->next->prev = o; \
}
    
#define SCHEDULE(o)	\
{ \
	if ((o) && !VISITED(o)) \
    { \
        if ((o)->prev) \
            (o)->prev->next = (o)->next; \
        else \
            unreachedObjects = (o)->next; \
        if ((o)->next) \
            (o)->next->prev = (o)->prev; \
        \
		(o)->next = scheduledObjects; \
		scheduledObjects = (ObjectRef)(o); \
		VISIT((o)); \
	} \
}

bool isInLinkedList(ObjectRef object, ObjectRef list)
{
	while (list)
	{
		if (object == list)
			return true;
		list = list->next;
	}
	return false;
}

void collectGarbage(MachineRef machine, CallObjectRef call, ObjectRef* stackTop, bool checkConsistency)
{
    assert(call != NULL);
	if (call)
	{
		assert(call->stackTop == NULL);
		call->stackTop = stackTop;
	}
    	
    ObjectRef unreachedObjects = machine->objects;
	ObjectRef scheduledObjects = NULL;
    ObjectRef reachedObjects = NULL;
    
	ObjectRef o = NULL;
	ObjectRef* stackItem;
	ULONG i;

	//if (checkConsistency)
	//{
	//	ObjectRef list = machine->objects;
	//	while (list)
	//	{
	//		SET_DEBUG(list);
	//		list = list->next;
	//	}
	//}
    
    SCHEDULE(machine->objectResult);
    SCHEDULE(machine->objectParameter);
	SCHEDULE(machine->objectNil);
	SCHEDULE(machine->objectYes);
	SCHEDULE(machine->objectNo);
    SCHEDULE(machine->objectPrototype);
    SCHEDULE(machine->objectRealPrototype);
    SCHEDULE(machine->objectFunctionPrototype);
    SCHEDULE(machine->objectNativeFunctionPrototype);
    SCHEDULE(machine->objectNilPrototype);
    SCHEDULE(machine->objectPropertyPrototype);
    SCHEDULE(machine->objectBoolPrototype);
    SCHEDULE(machine->objectStringPrototype);
    SCHEDULE(machine->objectArrayPrototype);
    SCHEDULE(machine->objectSequencePrototype);
    SCHEDULE(machine->objectClassPrototype);
    SCHEDULE(machine->objectInternalError);
    SCHEDULE(machine->objectDataPrototype);
   	SCHEDULE((ObjectRef)call);
    
	while (scheduledObjects)
	{
		DESCHEDULE(o); // take object from scheduled objects to process

		//if (checkConsistency && !DEBUGGED(o))
		//{
		//	((ObjectRef)NULL)->parent = NULL;
		//	wprintf(L"Pruser!\n");
		//}
		        
		SCHEDULE(o->parent);
		// Schedule all slots
		for (i=0; i<o->slotCount; i++)
			SCHEDULE(o->slots[i].value);
		
		// Schedule type-specific object references
		switch (o->type)
		{
			case typeFunction:
				SCHEDULE(FUNCTION(o)->parentCall);
				SCHEDULE(FUNCTION(o)->frozenCall);
				break;
            case typeBigCall:
			case typeCall:
				SCHEDULE(CALL(o)->parentCall);
				SCHEDULE(CALL(o)->calledFunction);
				SCHEDULE(CALL(o)->self);
				if (CALL(o)->stackTop)
				{
					stackItem = &CALL(o)->stack[0]; 
					while (stackItem <= CALL(o)->stackTop)
					{
						SCHEDULE(*stackItem);
						++stackItem;
					}
				}
				break;
			case typeArray:
				for (i=0; i<ARRAY(o)->size; i++)
					SCHEDULE(ARRAY(o)->data[i]);
				break;
            case typeClass:
                SCHEDULE(CLASS(o)->instancePrototype);
                break;
			default:
				break;
		}		
	}
    
	machine->objects = reachedObjects;
    while (reachedObjects)
    {
        UNVISIT(reachedObjects);
        reachedObjects = reachedObjects->next;
    }
    
	// Clean up the garbage
	ObjectRef garbage;
	while (unreachedObjects)
	{
		garbage = unreachedObjects;
        
		unreachedObjects = unreachedObjects->next;
        if (CACHED(garbage)) // if garbage object is a cached object, lets insert it to regular objects again instead of freeing it
        {
            if (machine->objects)
                machine->objects->prev = garbage;
            garbage->prev = NULL;
            garbage->next = machine->objects;
            machine->objects = garbage;
        }
        else
        {
            if (garbage->type != typeCall)
                freeObjectToPool(machine, garbage);
            else
                freeCallObjectToPool(machine, (CallObjectRef)garbage);
        }
	}
    
	if (call)
		call->stackTop = NULL;
}




#define TOP			   (*stackTop)
#define STACK_OBJECT_1 (*stackTop)
#define STACK_OBJECT_2 (*(stackTop-1))
#define STACK_OBJECT_3 (*(stackTop-2))
#define PUSH(o)			{ *(++stackTop) = (ObjectRef)o; }
#define POP()			stackTop--;
#define POP2()			stackTop -= 2;
#define POP_N(n)		stackTop -= n;

#define SET_TOP(o)		{ *stackTop = (o); }
#define NEW_REAL(value) newRealObject(objects, pool, value, machine->objectRealPrototype)
#define NEW_FUNCTION()	newObjectFromPool(objects, pool, typeFunction, machine->objectFunctionPrototype, sizeof(FunctionObject))
#define NEW_NATIVE_FUNCTION(symbol) newNativeFunction(machine, objects, pool, symbol, machine->objectNativeFunctionPrototype)

#define NEW_PROPERTY(readSymbol, writeSymbol) newPropertyObject(objects, pool, readSymbol, writeSymbol, machine->objectPropertyPrototype)
#define NEW_STRING(data, length) newStringObject(objects, pool, data, length, machine->objectStringPrototype)
#define NEW_ARRAY_OBJECT(data, length) newArrayObject(objects, pool, data, length, machine->objectArrayPrototype)
#define NEW_CLASS(instancePrototype, parent) newClassObject(objects, pool, instancePrototype, parent)

#define TOP_CALL		call

#define IS_YIELDED_FUNCTION(f)	(FUNCTION(f)->frozenCall != NULL)

#define PUSH_NEW_CALL(macroSelf, macroCalledFunction, macroType, macroSlotCount, macroStackLength, macroCallParent, macroLexicalParent) \
{ \
	CallObjectRef newCall = newCallObject(macroSlotCount, macroStackLength, objects, pool, callPools, (ObjectRef)macroLexicalParent); \
    newCall->self = macroSelf; \
	newCall->calledFunction = macroCalledFunction; \
	newCall->callType = macroType; \
	newCall->parentCall = macroCallParent; \
	newCall->stackTop = NULL; \
	newCall->bytecodeAddress = NULL; \
    newCall->yieldedFromCall = NULL; \
    newCall->errorBytecode = errorBytecode; \
	call = newCall;	\
}

#define POP_CALL() (call->stackTop = NULL, call = call->parentCall)

#define PERIODIC_COUNTER_INTERVAL 4000000//400000 // crash 400000 //2000 //4000000//4000000
 
#define STACK_LENGTH (stackTop - &call->stack[0] + 1)

#define DUMP_STACK_TRACE \
    CallObjectRef c = call; \
    wstring e; \
    LONG stackCounter = 50; \
	/* We want to ignore the outermost call, that is why c->parentCall != NULL... */ \
    while (c != NULL && c->parentCall != NULL && stackCounter-- > 0) \
    { \
        BytecodeInfo info; \
        infoForBytecode(machine->bytecodeInfos, machine->bytecodeInfoCount, c->errorBytecode - machine->bytecodes, &info); \
		e = wstring(L"[CALL] ") + machine->sourceIdTable[info.position.sourceIdIndex] + \
			L" " + info.position.toMinString() + L"\n" + e; \
        c = c->parentCall; \
    } \
    e += L"\n"; \
	wprintf(L"Error Stack Trace:\n"); \
    printString(e);

#define ERROR(text) \
{ \
	BytecodeInfo info; \
    DUMP_STACK_TRACE; \
	infoForBytecode(machine->bytecodeInfos, machine->bytecodeInfoCount, bytecodes - machine->bytecodes, &info); \
    wstring description = wstring(L"[RUNTIME ERROR] ") + \
        machine->sourceIdTable[info.position.sourceIdIndex] + \
        L" " + info.position.toMinString() + L"\n  " + text + L"\n"; \
    machine->position = info.position; \
    machine->error = errorNew(description, info.position); \
	printString(description); \
	return NULL; \
}

#define NATIVE_ERROR() ERROR(machine->internalErrorMessage)

#define ERROR3(text1, text2, text3) ERROR(text1 + text2 + text3)

#define MESSAGE(text) \
{ \
	BytecodeInfo info; \
	infoForBytecode(machine->bytecodeInfos, machine->bytecodeInfoCount, bytecodes - machine->bytecodes, &info); \
	printString(wstring() + machine->sourceIdTable[info.position.sourceIdIndex] + \
		info.position.toShortString() + L"  " text L"\n"); \
}

#define IS_NOT_KIND_OF_FUNCTION(object) ((object)->type != typeFunction && (object)->type != typeNativeFunction)


void hitStack(MachineRef machine, CallObjectRef call)
{
    unregisterHits(machine->bytecodeInfos, machine->bytecodeInfoCount);
    
    CallObjectRef c = call;
    while (c != NULL)
    {
        hitBytecode(machine->bytecodeInfos, machine->bytecodeInfoCount, c->errorBytecode - machine->bytecodes);
        c = c->parentCall;
    }
}

//#define PROFILE 1
#define PROFILER_FREQ 100000

#define TIME_LIMIT 1

ObjectRef machineRun(MachineRef machine, ObjectRef parameter)
{
	// Platform porting sanity check
	assert(sizeof(Slot) % sizeof(ObjectRef) == 0);

    machine->objectParameter = (parameter == NULL) ? machine->objectNil : parameter;
    BYTE* bytecodes = machine->bytecodes;
	ObjectRef* stackTop = NULL; // = &stack[-1];
	CallObjectRef call = NULL;
	ObjectRef* objects = &machine->objects;
	ObjectRef* pool = machine->pool;
    
	CallObjectRef (*callPools)[MAX_CALL_POOL_DATA_SIZE] = &machine->callPools;
	ObjectRef result = NULL;
	ULONG i, periodicCounter = PERIODIC_COUNTER_INTERVAL;
	ULONG symbol;
	ULONG length;
    ULONG profilerHits = 0;
	ULONG badInstructionCounter = 0;
    clock_t lastProfileHit = 0;
	
    BYTE* errorBytecode = NULL;
    
	// Call initializers
	CallType callType;
	BYTE callNumberOfParameters;
	BYTE* bytecodeContinuationAddress;
	bool isTailCall;
	
    CallObjectRef frozenCall = NULL;
    
	assert(*bytecodes == bINITIALIZE);
	
	#ifdef BENCHMARK
	clock_t time = clock();
	for (i=0; i<1; i++) 
	{
	bytecodes = machine->bytecodes;		
	call = NULL;
	#endif

        
    clock_t limit = clock();
 	while (true)
	{   

        #ifdef PROFILE
        clock_t now = clock();
        if (now - lastProfileHit > PROFILER_FREQ)
        {
            hitStack(machine, call);
            lastProfileHit = now;
            profilerHits++;
        }
        #endif

		if (!--periodicCounter)
		{
            if (haltSignal)
            {
                haltSignal = false;
                break;
            }

			collectGarbage(machine, call, stackTop, true);	
			periodicCounter = PERIODIC_COUNTER_INTERVAL;
            
#ifdef TIME_LIMIT
            if (clock()-limit > 3600000000)
            {
                ERROR(L"TIME LIMIT EXCEEDED! (error               )");
             //   wprintf(L"TIME LIMIT EXCEEDED! (error               )\n");
                break;
            }
#endif
		}

		// Stack Capacity Check - Activate for obscure memory bugs
//		if (call)
//			assert(stackTop - &call->stack[0] + 1 <= call->stackCapacity);



		//BytecodeInfo info; 
//                        DUMP_STACK_TRACE; 
        //infoForBytecode(machine->bytecodeInfos, machine->bytecodeInfoCount, bytecodes - machine->bytecodes, &info); 
		//wprintf(L"Line: %ls %d\n", machine->sourceIdTable[info.position.sourceIdIndex], info.position.minLine);
		
		switch (*bytecodes)
		{
			case bNUMBER:
            {
				PUSH(NEW_REAL(*(DOUBLE*)(bytecodes+1)));
                *bytecodes = bCACHEDNUMBER; //
				bytecodes += 1;
				*(ObjectRef*)bytecodes = *stackTop; //
                
                
				CACHE(*stackTop); //
				bytecodes += sizeof(DOUBLE);
				continue;
            }
			case bCACHEDNUMBER:
            {
				result = *(ObjectRef*)(bytecodes+1);
				PUSH(*(ObjectRef*)(bytecodes+1));
				bytecodes += 1 + sizeof(DOUBLE);
				continue;
            }
			case bSTRING:
            {
				length = *(ULONG*)(bytecodes+1);
				bytecodes += 1 + sizeof(ULONG);

				CHAR* s = NULL;
				ULONG stringSize = length / sizeof(CHAR32);
				NEW_ARRAY(s, CHAR, stringSize);
				for (ULONG i=0; i<stringSize; i++)
					s[i] = *(((CHAR32*)bytecodes) + i);


				PUSH(NEW_STRING(s, stringSize));

				free(s);

				bytecodes += length;
				continue;
            }
			case bLIST:
            {
				length = *(ULONG*)(bytecodes+1);
				result = (ObjectRef)NEW_ARRAY_OBJECT(stackTop-length+1, length);
				POP_N(length);
				PUSH(result);
				bytecodes += 1 + sizeof(ULONG);
				continue;
            }
            case bNULL:
            {
                PUSH(NULL);
                bytecodes += 1;
                continue;
            }
			case bNIL:
            {
				PUSH(machine->objectNil);
				bytecodes += 1;
				continue;
            }
            case bSELF_FOR_CALL: // fall through
            {
				PUSH(NULL);
            }
			case bSELF:
            {
//                if (bytecodes - machine->bytecodes == 5838)
//                    wprintf(L"self");
				PUSH(call->self ? call->self : machine->objectNil);
				bytecodes += 1;
				continue;
            }
			case bPROPERTY:
            {
				PUSH(NEW_PROPERTY(*(ULONG*)(bytecodes+1), *(ULONG*)(bytecodes+1+sizeof(ULONG))));
				bytecodes += 1 + sizeof(ULONG) + sizeof(ULONG);
				continue;
			}	
			case bCAPTURE:
            {
				*bytecodes = bRESTORE;
				bytecodes += 1;
				*(ULONG*)bytecodes = STACK_LENGTH;
				bytecodes += sizeof(ULONG);
				continue;
			}	
			case bRESTORE_FROM: // fall through
            {
				*bytecodes = bRESTORE;
				*(ULONG*)(bytecodes+1) = *(bytecodes + *(LONG*)(bytecodes+1) + 1);
            }
			case bRESTORE:
            {
				stackTop = *(ULONG*)(bytecodes+1) + &call->stack[0] - 1;
				bytecodes += 1 + sizeof(ULONG);
				continue;
			}	
			case bPOP:
            {
				stackTop -= *(ULONG*)(bytecodes+1);
				bytecodes += 1 + sizeof(ULONG);
				continue;
            }
			case bPOP_NON_TOP:
            {
				ObjectRef* newStackTop = stackTop - *(ULONG*)(bytecodes+1);
				*newStackTop = *stackTop;
				stackTop = newStackTop;
				bytecodes += 1 + sizeof(ULONG);
				continue;
            }
			case bJUMP_FROZEN_POP:
            {
				if (TOP->type != typeFunction)
					ERROR(L"Property 'generator' of iterable object did not return a function.");
				if (FUNCTION(TOP)->frozenCall)
				{
					POP();
					bytecodes += 1 + sizeof(LONG);
				}
				else
				{
					bytecodes += *(LONG*)(bytecodes+1);
				}
				continue;
            }
			case bJUMPNO:
            {
				if (TOP == machine->objectNo)
					bytecodes += *(LONG*)(bytecodes+1);
				else
					bytecodes += 1 + sizeof(LONG);
				POP();
				continue;
            }
			case bJUMPNO_NIL:
            {
				if (TOP == machine->objectNo)
                {
					bytecodes += *(LONG*)(bytecodes+1);
                    TOP = machine->objectNil;
                }
				else
                {
					bytecodes += 1 + sizeof(LONG);
                    POP();
                }

				continue;
            }
			case bJUMP:
            {
				bytecodes += *(LONG*)(bytecodes+1);
				continue;
            }
                /* // Number are immutable and cached...
			case bDECREMENT:
            {
				REAL(STACK_OBJECT_2)->value -= REAL(STACK_OBJECT_1)->value;
				POP();
				bytecodes += 1;
				continue;
            }
			case bINCREMENT:
            {
				REAL(STACK_OBJECT_2)->value += REAL(STACK_OBJECT_1)->value;
				POP();
				bytecodes += 1;
				continue;
            }*/
            case bPLUS_CONST:
            {
                if (STACK_OBJECT_1->type == typeReal)
                {
                    DOUBLE sum = REAL(STACK_OBJECT_1)->value + *(DOUBLE*)(bytecodes+1);
                    POP();
                    PUSH(NEW_REAL(sum));
                    bytecodes += 1 + sizeof(DOUBLE);
                    continue;
                }
                errorBytecode = bytecodes;
                PUSH(NEW_REAL(*(DOUBLE*)(bytecodes+1)));
                bytecodes += sizeof(DOUBLE); 
                goto plusRealization;
                // fallthrough, the same situation as binary bytecode
            }
			case bPLUS: // currently only used as a support for bPLUS_CONST
            {
                errorBytecode = bytecodes;
            plusRealization:
                if (STACK_OBJECT_2->type == typeReal && STACK_OBJECT_1->type == typeReal)
                {
                    DOUBLE sum = REAL(STACK_OBJECT_2)->value + REAL(STACK_OBJECT_1)->value;
                    POP();
                    POP();
                    PUSH(NEW_REAL(sum));
                    bytecodes += 1;
                }
                else
                {
                    ObjectRef parameter = STACK_OBJECT_1;
                    ObjectRef method = getSlotRecursive(STACK_OBJECT_2, SYMBOL_PLUS);
                    if (!method)
                        ERROR3(L"Unknown method '", machine->symbolTable[SYMBOL_PLUS], L"' inside this object");
                    
                    POP();
                    
                    PUSH(method);
                    PUSH(parameter);
                    
                    callType = CALL_TYPE_RETURNABLE;
                    callNumberOfParameters = 1;
                    bytecodeContinuationAddress = bytecodes + 1;
                    isTailCall = false;
                    goto callRealization;
                }
				continue;
            }
			/*case bUNARYMINUS:
            {
				TOP = (ObjectRef)NEW_REAL(-REAL(TOP)->value);
				bytecodes += 1;
				continue;
            }
            case bUNARYPLUS:
            {
                bytecodes += 1;
                continue;
            }*/
            case bMINUS_CONST:
            {
                if (STACK_OBJECT_1->type == typeReal)
                {
                    DOUBLE sum = REAL(STACK_OBJECT_1)->value - *(DOUBLE*)(bytecodes+1);
                    POP();
                    PUSH(NEW_REAL(sum));
                    bytecodes += 1 + sizeof(DOUBLE);
                    continue;
                }
                errorBytecode = bytecodes;
                PUSH(NEW_REAL(*(DOUBLE*)(bytecodes+1)));
                bytecodes += sizeof(DOUBLE); 
                goto minusRealization;
                // fallthrough, the same situation as binary bytecode
            }
			case bMINUS: // currently only used as a support for bPLUS_CONST
            {
                errorBytecode = bytecodes;
            minusRealization:
                if (STACK_OBJECT_2->type == typeReal && STACK_OBJECT_1->type == typeReal)
                {
                    DOUBLE sum = REAL(STACK_OBJECT_2)->value - REAL(STACK_OBJECT_1)->value;
                    POP();
                    POP();
                    PUSH(NEW_REAL(sum));
                    bytecodes += 1;
                }
                else
                {
                    ObjectRef parameter = STACK_OBJECT_1;
                    ObjectRef method = getSlotRecursive(STACK_OBJECT_2, SYMBOL_MINUS);
                    if (!method)
                        ERROR3(L"Unknown method '", machine->symbolTable[SYMBOL_MINUS], L"' inside this object");
                    
                    POP();
                    
                    PUSH(method);
                    PUSH(parameter);
                    
                    callType = CALL_TYPE_RETURNABLE;
                    callNumberOfParameters = 1;
                    bytecodeContinuationAddress = bytecodes + 1;
                    isTailCall = false;
                    errorBytecode = bytecodes;
                    goto callRealization;
                }
				continue;
            }
			/*
			case bTIMES:
            {
				DOUBLE product = REAL(STACK_OBJECT_2)->value * REAL(STACK_OBJECT_1)->value;
				POP();
				POP();
				PUSH(NEW_REAL(product));
				bytecodes += 1;
				continue;
            }
			case bDIVIDE:
            {
				DOUBLE ratio = REAL(STACK_OBJECT_2)->value / REAL(STACK_OBJECT_1)->value;
				POP();
				POP();
				PUSH(NEW_REAL(ratio));
				bytecodes += 1;
				continue;
            }
			case bMOD:
            {
				DOUBLE ratio = (ULONG)REAL(STACK_OBJECT_2)->value % (ULONG)REAL(STACK_OBJECT_1)->value;
				POP();
				POP();
				PUSH(NEW_REAL(ratio));
				bytecodes += 1;
				continue;
            }
			case bDIV:
            {
				DOUBLE ratio = (ULONG)REAL(STACK_OBJECT_2)->value / (ULONG)REAL(STACK_OBJECT_1)->value;
				POP();
				POP();
				PUSH(NEW_REAL(ratio));
				bytecodes += 1;
				continue;
            }
			case bEXP:
            {
				DOUBLE ratio = pow(REAL(STACK_OBJECT_2)->value, REAL(STACK_OBJECT_1)->value);
				POP();
				POP();
				PUSH(NEW_REAL(ratio));
				bytecodes += 1;
				continue;
            }*/

			case bNOT:
            {
				result = TOP == machine->objectYes ? machine->objectNo : machine->objectYes;
				POP();
				PUSH(result);
				bytecodes += 1;
				continue;
            }

			/*
			case bLESS:
            {
				result = REAL(STACK_OBJECT_2)->value < REAL(STACK_OBJECT_1)->value ? machine->objectYes : machine->objectNo;
				POP();
				POP();
				PUSH(result);
				bytecodes += 1;
				continue;
            }
			case bLESSOREQUAL:
            {
				result = REAL(STACK_OBJECT_2)->value <= REAL(STACK_OBJECT_1)->value ? machine->objectYes : machine->objectNo;
				POP();
				POP();
				PUSH(result);
				bytecodes += 1;
				continue;
            }
            case bMOREOREQUAL:
            {
				result = REAL(STACK_OBJECT_2)->value >= REAL(STACK_OBJECT_1)->value ? machine->objectYes : machine->objectNo;
				POP();
				POP();
				PUSH(result);
				bytecodes += 1;
				continue;
            }
			case bMORE:
            {
				result = REAL(STACK_OBJECT_2)->value > REAL(STACK_OBJECT_1)->value ? machine->objectYes : machine->objectNo;
				POP();
				POP();
				PUSH(result);
				bytecodes += 1;
				continue;
            }
            case bAND:
            {
				result = BOOL(STACK_OBJECT_2)->value && BOOL(STACK_OBJECT_1)->value ? machine->objectYes : machine->objectNo;
				POP();
				POP();
				PUSH(result);
				bytecodes += 1;
				continue;
            }
            case bOR:
            {
				result = BOOL(STACK_OBJECT_2)->value || BOOL(STACK_OBJECT_1)->value ? machine->objectYes : machine->objectNo;
				POP();
				POP();
				PUSH(result);
				bytecodes += 1;
				continue;
            }*/
			case bVAR:
            {
				addSlotToCall(call, *(ULONG*)(bytecodes+1), machine->objectNil);
				PUSH(machine->objectNil);
				bytecodes += 1 + sizeof(ULONG);
				continue;
            }
			case bNATIVE:
            {
				symbol = *(ULONG*)(bytecodes+1);
				result = (ObjectRef)NEW_NATIVE_FUNCTION(symbol);
				if (!result)
				{
					errorBytecode = bytecodes;
					ERROR3(L"Unknown native function '", machine->symbolTable[*(ULONG*)(bytecodes+1)], L"'");
				}

				addSlotToCall(call, symbol, result);
				PUSH(result);				
				bytecodes += 1 + sizeof(ULONG);
				continue;
			}	
			case bSLOT:
            {
				addSlot(call->self, *(ULONG*)(bytecodes+1), machine->objectNil);
				PUSH(machine->objectNil);
//				printObject(machine, call->call.self);
				bytecodes += 1 + sizeof(ULONG);
				continue;
			}	
			case bASSIGN_SLOT:
            {
				if (!setSlot(call->self, *(ULONG*)(bytecodes+1), STACK_OBJECT_1))
				{
					errorBytecode = bytecodes;
					ERROR3(L"Unknown identifier '", machine->symbolTable[*(ULONG*)(bytecodes+1)], L"'");
				}

				bytecodes += 1 + sizeof(ULONG);
				continue;
            }
			case bASSIGNIN:
            {
                symbol = *(ULONG*)(bytecodes+1);
				ObjectRef receiver = STACK_OBJECT_2;
				if ((result = getSlotRecursive(receiver, symbol)))
				{
					if (result->type != _typeProperty)
					{
                        setSlotRecursive(STACK_OBJECT_2, *(ULONG*)(bytecodes+1), STACK_OBJECT_1);
                        STACK_OBJECT_2 = STACK_OBJECT_1;
                        POP();
					}
					else
					{
                        ObjectRef parameter = TOP;
                        POP();
						ObjectRef writeFunction = getSlotRecursive(receiver, PROPERTY(result)->writeSymbol);
						PUSH(writeFunction);
                        PUSH(parameter);
						
						callType = CALL_TYPE_RETURNABLE;
						callNumberOfParameters = 1;
						bytecodeContinuationAddress = bytecodes + 1 + sizeof(ULONG);
						isTailCall = false;
                        errorBytecode = bytecodes;
						goto callRealization;
					}
				}
				else
				{
					errorBytecode = bytecodes;
					ERROR3(L"Unknown slot '", machine->symbolTable[symbol], L"' inside other object");
				}
                                
//				if (setSlotRecursive(STACK_OBJECT_2, *(ULONG*)(bytecodes+1), STACK_OBJECT_1))
//				{
//					STACK_OBJECT_2 = STACK_OBJECT_1;
//					POP();
//				}
//				else
//					ERROR3(L"Unknown slot '", machine->symbolTable[*(ULONG*)(bytecodes+1)], L"' inside other object");
				
				bytecodes += 1 + sizeof(ULONG);
				continue;
			}	
			case bASSIGN:
            {
				symbol = *(ULONG*)(bytecodes+1);
				if (!setSlot((ObjectRef)call, symbol, STACK_OBJECT_1) &&
					!setSlotRecursive(call->self, symbol, STACK_OBJECT_1) &&
					!setCallSlotRecursive((CallObjectRef)(call->parent), symbol, STACK_OBJECT_1))
				{
					errorBytecode = bytecodes;
					ERROR3(L"Unknown identifier '", machine->symbolTable[symbol], L"'");
				}
				bytecodes += 1 + sizeof(ULONG);
				continue;
			}	
//			case bEQUAL:
//            {
//				if (STACK_OBJECT_1->type == STACK_OBJECT_2->type)
//				{
//					if (STACK_OBJECT_1->type == typeReal)
//						result = REAL(STACK_OBJECT_1)->value == REAL(STACK_OBJECT_2)->value ? machine->objectYes : machine->objectNo;
//					else if (STACK_OBJECT_1->type == typeString)
//                        result = STRING(STACK_OBJECT_1)->size == STRING(STACK_OBJECT_2)->size && wcscmp(STRING(STACK_OBJECT_1)->data, STRING(STACK_OBJECT_2)->data) == 0 ? machine->objectYes : machine->objectNo;
//                    else
//						result = STACK_OBJECT_1 == STACK_OBJECT_2 ? machine->objectYes : machine->objectNo;
//				}
//				else
//					result = machine->objectNo;
//				POP();
//				POP();
//				PUSH(result);
//				bytecodes += 1;
//				continue;
//			}
            
			case bFINDIN_FOR_CALL:
			case bFINDIN:
            {
				symbol = *(ULONG*)(bytecodes+1);
				ObjectRef receiver = STACK_OBJECT_1;
				if ((result = getSlotRecursive(receiver, symbol)))
				{
					if (*bytecodes != bFINDIN_FOR_CALL)
						POP();
					if (result->type != _typeProperty)
					{
						PUSH(result)
					}
					else
					{
						ObjectRef readFunction = getSlotRecursive(receiver, PROPERTY(result)->readSymbol);
						PUSH(receiver);
						PUSH(readFunction);
						
						callType = CALL_TYPE_RETURNABLE;
						callNumberOfParameters = 0;
						bytecodeContinuationAddress = bytecodes + 1 + sizeof(ULONG);
						isTailCall = false;
                        errorBytecode = bytecodes;
						goto callRealization;
					}
				}
				else
				{
					errorBytecode = bytecodes;
					ERROR3(L"Unknown slot '", machine->symbolTable[symbol], L"' inside other object");
				}
				bytecodes += 1 + sizeof(ULONG);
				continue;
			}		
			case bFIND_FOR_CALL:
			case bFIND:
			{
				symbol = *(ULONG*)(bytecodes+1);
				bool isMethod = false;
				

				if ((result = getCallSlotRecursive((CallObjectRef)call, symbol)))
				{
					if (*bytecodes == bFIND_FOR_CALL)
						PUSH(isMethod ? call->self : NULL);
					
                    PUSH(result);

					//if (badInstructionCounter == 13588393)
					//{
					//	bool isInList = isInLinkedList((ObjectRef)(call->parent), machine->objects);

					//	((ObjectRef)isInList)->tag = 123;
					//	wprintf(L"Is in list: %d\n", isInList);
					//}
				}
				else
                {

                    errorBytecode = bytecodes;
					ERROR3(L"Unknown identifier '", machine->symbolTable[symbol], L"'");
                }
				bytecodes += 1 + sizeof(ULONG);
				continue;
			}
			case bBEGIN:
            {
				result = NEW_FUNCTION();
				ULONG slotCount = *(ULONG*)(bytecodes+1);
				bytecodes += 1 + sizeof(ULONG);
				
				ULONG stackSize = *(ULONG*)(bytecodes);
				bytecodes += sizeof(ULONG);

				BYTE* endBytecodeAddress = machine->bytecodes + *(ULONG*)(bytecodes);
				bytecodes += sizeof(ULONG);
				
				BYTE numberOfParameters = *bytecodes;
				bytecodes += 1;
				FUNCTION(result)->numberOfSlotsNeededForCall = slotCount;
				FUNCTION(result)->stackSizeNeeded = stackSize;
				FUNCTION(result)->numberOfParameters = numberOfParameters;
				FUNCTION(result)->parameters = (ULONG*)bytecodes;
				FUNCTION(result)->startAddress = bytecodes + numberOfParameters*sizeof(ULONG);
				FUNCTION(result)->parentCall = call;
				FUNCTION(result)->frozenCall = NULL;
//				result->function.self = call->call.self;
				bytecodes = endBytecodeAddress + 1; // Skip end - maybe precompute skipped?
				PUSH(result);
				continue;
            }
			case bTAILCALL:
			case bCALL:
            {
				callType = (CallType)*(bytecodes + 1);
                assert(callType == CALL_TYPE_RETURNABLE || callType == CALL_TYPE_TRANSPARENT);
				callNumberOfParameters = *(bytecodes + 1 + 1); 
				bytecodeContinuationAddress = bytecodes + 1 + 1 + 1;
				isTailCall = *bytecodes == bTAILCALL && TAIL_CALL_OPTIMIZATION_ON;
                errorBytecode = bytecodes;
				
			callRealization:
				ObjectRef* parameters = stackTop - callNumberOfParameters + 1; 
				FunctionObjectRef functionObject = (FunctionObjectRef)*(parameters-1);
				ObjectRef self = *(parameters-2);
                
                if (IS_NOT_KIND_OF_FUNCTION(functionObject))
                {
                    self = (ObjectRef)functionObject;
                    functionObject = (FunctionObjectRef)getSlotRecursive(self, SYMBOL_call);
                    if (functionObject == NULL || IS_NOT_KIND_OF_FUNCTION(functionObject))
                    {
                        BytecodeInfo info; 
//                        DUMP_STACK_TRACE; 
                        infoForBytecode(machine->bytecodeInfos, machine->bytecodeInfoCount, bytecodes - machine->bytecodes, &info); 
                        wstring description = wstring(L"This is not a function and can not be called1., ");
                        description += info.position.toShortString();
                        
                        machine->position = info.position; 
                        machine->error = errorNew(description, info.position); 
                        return NULL; 
                        
//                        ERROR(L"This is not a function and can not be called1.");
                    }
                }

				if (callNumberOfParameters > functionObject->numberOfParameters) 
				{ 
                    ERROR(L"Too many parameters");
				} 
				while (callNumberOfParameters < functionObject->numberOfParameters) 
				{ 
					PUSH(machine->objectNil); 
					++callNumberOfParameters; 
				} 
				assert(functionObject->numberOfParameters == callNumberOfParameters); 
				
				stackTop = parameters - 3;
				
				if (functionObject->type == typeNativeFunction)
				{
					//wprintf(L"Calling native function\n");
					PUSH(NATIVE_FUNCTION(functionObject)->address(parameters-2, machine));
					//wprintf(L"Done native function\n");
                    if (TOP == machine->objectInternalError)
                    {
                        NATIVE_ERROR();
                    }
//					assert(*bytecodeContinuationAddress);
					bytecodes = bytecodeContinuationAddress;
					continue;
				}
				else if (functionObject->type == typeFunction)
				{
					if (self == NULL)
						self = *(parameters-2) = functionObject->parentCall->self;
				}
				else
				{
					ERROR(L"This is not a function and can not be called2.");
				}
				
				if (isTailCall) // tail calls
				{
					while (call->callType != CALL_TYPE_RETURNABLE)
						POP_CALL();
                    errorBytecode = call->errorBytecode;
					POP_CALL();
				}
				else // normal calls
				{
					// freeze
					call->stackTop = stackTop;
					call->bytecodeAddress = bytecodeContinuationAddress;
				}
				
				if IS_YIELDED_FUNCTION(functionObject)
				{
//					printObject(machine, (ObjectRef)functionObject);
					CallObjectRef yieldedCall = functionObject->frozenCall;
					functionObject->frozenCall = NULL;
					
//					yieldedCall->parentCall = call;
                    yieldedCall->yieldedFromCall->parentCall = call;
					yieldedCall->self = self;
					yieldedCall->callType = callType;
                    yieldedCall->errorBytecode = errorBytecode;
					
					// unfreeze call
					bytecodes = yieldedCall->bytecodeAddress;
					stackTop = yieldedCall->stackTop;
					call = yieldedCall;
					
					BYTE i;
					for (i=0; i < callNumberOfParameters; i++)
						if (!setSlot((ObjectRef)call, functionObject->parameters[i], *(parameters++)))
						{
							wprintf(L"%s:%u: setting parameter in yielded function that does not exist!\n", __FILE__, __LINE__);
							abort();
						}
				}
				else
				{
					PUSH_NEW_CALL(self,
						functionObject,
						callType,
						functionObject->numberOfSlotsNeededForCall, 
						functionObject->stackSizeNeeded, 
						call,
						functionObject->parentCall);
                    
					bytecodes = functionObject->startAddress;
					stackTop = &call->stack[-1];
                    
					BYTE i; 
					for (i=0; i < callNumberOfParameters; i++) 
						addSlotToCall(call, functionObject->parameters[i], *(parameters++)); 
				}
				call->stackTop = NULL;
				continue;
            }
            case bRETURNFROM: // fall trough
            {
                while (call && (ObjectRef)call->calledFunction != TOP)
					POP_CALL();
                if (!call)
                    ERROR(L"Cannot return from this object, it is not being called.");
                POP();
            }
			case bRETURN:
            {
                assert(call->callType == CALL_TYPE_RETURNABLE);
				while (call->callType != CALL_TYPE_RETURNABLE)
					POP_CALL();
				
				call->calledFunction->frozenCall = NULL;
				
				// unfreeze parent
				result = *stackTop;
				bytecodes = call->parentCall->bytecodeAddress;
				stackTop = call->parentCall->stackTop;
				call = call->parentCall;
				PUSH(result);
				
				call->stackTop = NULL;
				continue;
            }
            case bYIELDFROM:
            {
                CallObjectRef frozenCall = call;
                while (call && (ObjectRef)call->calledFunction != TOP)
                {
                    call = call->parentCall;
                }
                if (!call)
                {
                    ERROR(L"Cannot yield from this object, it is not being called (maybe it was optimized away as tail recursion?).");
                }
                POP();
                
				call->calledFunction->frozenCall = frozenCall;
                
				frozenCall->bytecodeAddress = bytecodes + 1;
				frozenCall->stackTop = stackTop;
                frozenCall->yieldedFromCall = call;
				
				// unfreeze parent
				result = *stackTop;
				bytecodes = call->parentCall->bytecodeAddress;
				stackTop = call->parentCall->stackTop;
				call = call->parentCall;
				PUSH(result);
				
				call->stackTop = NULL;
				continue;
            }
			case bYIELD:
            {
                CallObjectRef frozenCall = call;
				while (call->callType != CALL_TYPE_RETURNABLE)
                    call = call->parentCall;

				call->calledFunction->frozenCall = frozenCall;
				frozenCall->bytecodeAddress = bytecodes + 1;
				frozenCall->stackTop = stackTop;
                frozenCall->yieldedFromCall = call;
				
				// unfreeze parent
				result = *stackTop;
				bytecodes = call->parentCall->bytecodeAddress;
				stackTop = call->parentCall->stackTop;
				call = call->parentCall;
				PUSH(result);
				
				call->stackTop = NULL;
				continue;
            }
			case bEND:
            {
				call->calledFunction->frozenCall = NULL;
				
				// unfreeze parent
				result = *stackTop;
				bytecodes = call->parentCall->bytecodeAddress;
				stackTop = call->parentCall->stackTop;
				call = call->parentCall;
				PUSH(result);
				
				call->stackTop = NULL;
				continue;
            }
			case bASSERT:
            {
				if (STACK_OBJECT_1 != machine->objectYes)
					ERROR(L"!!! ASSERT FAILED !!!")
				else
					MESSAGE(L"Assert OK.");
				
				bytecodes += 1;
				continue;
            }
			case bPRINT:
            {
				wprintf(L"RESULT:\n");
				printObject(machine, TOP);
				bytecodes += 1;
				continue;
            }
			case bINITIALIZE:
            {
                errorBytecode = bytecodes;
				PUSH_NEW_CALL(NULL, NULL, CALL_TYPE_RETURNABLE, *(ULONG*)(bytecodes+1), *(ULONG*)(bytecodes+1+sizeof(ULONG)), NULL, NULL);
				machine->objectResult = (ObjectRef)call;
				stackTop = &call->stack[-1];
				bytecodes += 1 + sizeof(ULONG) + sizeof(ULONG);
				continue;
            }
			case bDONE:
            {
				// POP_CALL();
				goto exit;
				continue;
            }
			default:
            {
				wprintf(L"Unknown bytecode: %d at position: %d\n", *bytecodes, bytecodes - machine->bytecodes);
                
                BytecodeInfo info;
                infoForBytecode(machine->bytecodeInfos, machine->bytecodeInfoCount, bytecodes - machine->bytecodes, &info);
                wstring description(L"Unknown bytecode, ");
                description += info.position.toShortString();
                machine->position = info.position;
                machine->error = errorNew(description, info.position);
                
				return NULL;
            }
		}

	}
exit:;
    collectGarbage(machine, call, stackTop, false);
//		printObjects(machine);

#ifdef BENCHMARK
	}
	wprintf(L"Elapsed time: %d", clock()-time);
#endif
    
#ifdef PROFILE
    printProfilerStatistics(machine->bytecodeInfos, machine->bytecodeInfoCount, profilerHits);
#endif
    
    return machine->objectResult;
}

MachineRef machineNew(BYTE* bytecodeData, BYTE* bytecodes, CHAR** symbolTable, ULONG symbolTableLength, CHAR** sourceIdTable, ULONG sourceIdTableLength, BytecodeInfo* bytecodeInfos, ULONG bytecodeInfoCount)
{
    MachineRef machine = new Machine();
    
    
	ObjectRef* objects = &machine->objects;
	ObjectRef* pool = machine->pool;
    
    machine->bytecodeData = bytecodeData;
	machine->bytecodes = bytecodes;
	machine->symbolTable = symbolTable;
	machine->symbolTableLength = symbolTableLength;
    machine->sourceIdTable = sourceIdTable;
    machine->sourceIdTableLength = sourceIdTableLength;
	machine->bytecodeInfos = bytecodeInfos;
	machine->bytecodeInfoCount= bytecodeInfoCount;
	
	machine->objectPrototype = newObjectFromPool(objects, pool, typeObject, NULL, sizeof(Object));
	
	machine->objectClassPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectRealPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectFunctionPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectSequencePrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectStringPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectSequencePrototype, sizeof(Object));
	machine->objectNativeFunctionPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectNilPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectPropertyPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectBoolPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	machine->objectArrayPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectSequencePrototype, sizeof(Object));
	machine->objectDataPrototype = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
	
	machine->objectNil = newObjectFromPool(objects, pool, typeNil, machine->objectNilPrototype, sizeof(Object));
	machine->objectYes = (ObjectRef)newBoolObject(objects, pool, true, machine->objectBoolPrototype);
	machine->objectNo = (ObjectRef)newBoolObject(objects, pool, false, machine->objectBoolPrototype);
    
    machine->objectInternalError = newObjectFromPool(objects, pool, typeInternalPrototype, machine->objectPrototype, sizeof(Object));
    machine->objectResult = NULL;
    machine->objectParameter = machine->objectNil;
    
    return machine;
}

#define VERSION_STRING_SIZE 20

MachineRef machineNew(BYTE* bytes, ULONG byteCount)
{

    BYTE* bytecodeData;
    NEW_ARRAY(bytecodeData, BYTE, byteCount);
    memcpy(&bytecodeData[0], &bytes[0], sizeof(BYTE)*byteCount);
    
    bytes = bytecodeData;
	BYTE* obytes = bytes;

    CHAR32 version[VERSION_STRING_SIZE+1];
    memcpy(&version[0], &bytes[0], sizeof(CHAR32) * VERSION_STRING_SIZE);
    version[VERSION_STRING_SIZE] = NULL;
    
//    wprintf(L"Compiler version: %ls\n", bytes);
//    wprintf(L"--------------------------\n");
    
    bytes += VERSION_STRING_SIZE * sizeof(CHAR32);
    
    ULONG bytecodeSize = *(ULONG*)bytes;
    bytes += sizeof(ULONG);
    
    BYTE* bytecodes = bytes;
    
    bytes += bytecodeSize;
    
	//wprintf(L"Ulong size: %d\n", sizeof(ULONG));
	//wprintf(L"Double size: %d\n", sizeof(DOUBLE));
    
    ULONG symbolTableLength = *(ULONG*)bytes;
    bytes += sizeof(ULONG);
    
    CHAR** symbolTable = NULL;
    NEW_ARRAY(symbolTable, CHAR*, symbolTableLength);
    
    for (ULONG i=0; i < symbolTableLength; i++)
    {
        ULONG stringLengthIncludingNull = *(ULONG*)bytes;
        bytes += sizeof(ULONG);

		NEW_ARRAY(symbolTable[i], CHAR, stringLengthIncludingNull / sizeof(CHAR32));
		for (ULONG j=0; j < stringLengthIncludingNull / sizeof(CHAR32); j++)
		{
			CHAR32 bigChar = *(((CHAR32*)bytes) + j);
			symbolTable[i][j] = bigChar;
		}

        //symbolTable[i] = (CHAR*)bytes;
        bytes += stringLengthIncludingNull;

//        NSString* s = [[[NSString alloc] initWithBytes: symbolTable[i] length: sizeof(CHAR)*(stringLengthIncludingNull-1) encoding: NSUTF32LittleEndianStringEncoding] autorelease];
//        NSLog(@"$ %@", s);
    }

    
    ULONG sourceIdTableLength = *(ULONG*)bytes;
    bytes += sizeof(ULONG);
    
    CHAR** sourceIdTable = NULL;
    NEW_ARRAY(sourceIdTable, CHAR*, sourceIdTableLength);
    
    for (ULONG i=0; i < sourceIdTableLength; i++)
    {
        ULONG stringLengthIncludingNull = *(ULONG*)bytes;
        bytes += sizeof(ULONG);

		NEW_ARRAY(sourceIdTable[i], CHAR, stringLengthIncludingNull / sizeof(CHAR32));
		for (ULONG j=0; j < stringLengthIncludingNull / sizeof(CHAR32); j++)
		{
			CHAR32 bigChar = *(((CHAR32*)bytes) + j);
			sourceIdTable[i][j] = bigChar;
		}

        //sourceIdTable[i] = (CHAR*)bytes;
        bytes += stringLengthIncludingNull;
        
//        NSString* s = [[[NSString alloc] initWithBytes: symbolTable[i] length: sizeof(CHAR)*(stringLengthIncludingNull-1) encoding: NSUTF32LittleEndianStringEncoding] autorelease];
        //        NSLog(@"$ %@", s);
    }
    
    
    
    assert(sizeof(BytecodeInfo) == 72);
    ULONG bytecodeInfoCount = *(ULONG*)bytes;
    bytes += sizeof(ULONG);
    BytecodeInfo* bytecodeInfos = (BytecodeInfo*)bytes;
    
    return machineNew(bytecodeData, bytecodes, symbolTable, symbolTableLength, sourceIdTable, sourceIdTableLength, bytecodeInfos, bytecodeInfoCount);
}

MachineRef machineNew(const char* path)
{
    FILE* file = fopen(path, "rb");
    if (file == NULL)
    {
        wprintf(L"Could not open ycode file: %s\n", path);
        return NULL;
    }
    
    fseek(file, 0, SEEK_END);
    ULONG byteCount = ftell(file);
    fseek(file, 0, SEEK_SET);
    
    BYTE* bytes = (BYTE*)malloc(byteCount);
    fread(bytes, byteCount, 1, file);
    fclose(file);
    
    MachineRef machine = machineNew(bytes, byteCount);
    
    free(bytes);
    
    return machine;
}

MachineRef machineNew(CompilerRef compiler)
{
    BYTE* data = NULL;
    ULONG dataSize = 0;
    if (!compilerGetResult(compiler, &data, &dataSize))
        return NULL;
    return machineNew(data, dataSize);
}

Machine::~Machine()
{
    free(bytecodeData);

    ULONG i;

//	for (i=0; i<machine->symbolTableLength; i++)
//		free(machine->symbolTable[i]);
//	free(machine->symbolTable);

    delete error;
    
    while (objects)
    {
        ObjectRef o = objects;
        objects = o->next;
        
        if (o->type != typeCall)
            freeObjectToPool(this, o);
        else
            freeCallObjectToPool(this, (CallObjectRef)o);
    }
	
    for (i=0; i < numberOfObjectTypes; i++)
    {
        ObjectRef _pool = pool[i];
        while (_pool)
        {
            ObjectRef o = _pool;
            _pool = o->next;
            free(o);
        }
    }
    
    for (i=0; i < MAX_CALL_POOL_DATA_SIZE; i++)
    {
        ObjectRef _pool = (ObjectRef)callPools[i];
        while (_pool)
        {
            ObjectRef o = _pool;
            _pool = o->next;
            free(o);
        }
    }
}
