#include <iostream>

#include "machine.h"


#include "object.h"

using std::string;

string readFile(const string& path)
{
	FILE* file = fopen(path.c_str(), "r");
	assert(file != NULL);

	string result;
	char buf[1024];
	size_t len;
	while ((len = fread(buf, 1, sizeof(buf), file)) != 0)
	{
		result.append(buf, len);
	}
	fclose(file);

	return result;
}

string removeSuffix(string path, string separators)
{
	size_t pos = path.find_last_of(separators);
	return (pos == string::npos) ? string() : path.substr(0, pos + 1);
}

bool endsWith(const string& haystack, const string& suffix)
{
	return (haystack.size() >= suffix.size() &&
		haystack.substr(haystack.size() - suffix.size()) == suffix);
}

wstring widen(const string& s)
{
	return wstring(s.begin(), s.end());
}

int main (int argc, char * const argv[]) 
{
	if (argc <= 1)
	{
		printf("Yon Virtual Machine\n");
		printf("*******************\n");
		printf("\n");
		printf("Compiling:           > yon.exe \"filePath.yon\"\n");
		printf("Compiling & Running: > yon.exe \"filePath.yon\" r\n");
		printf("Compiling Compiler:  > yon.exe \"filePath.yon\" c\n\n");
		printf("Running:             > yon.exe \"filePath.ycode\"\n");
		printf("\n");
		return 0;
	}

	string exePath(argv[0]);
	string filePath(argv[1]);
	string directoryPath = removeSuffix(filePath, "/\\");
	string libraryPath = removeSuffix(exePath, "/\\") + "library.yon";
	string compilerPath = removeSuffix(exePath, "/\\") + "compiler.ycode";

	
	bool needsCompilation = endsWith(filePath, ".yon");

	if (needsCompilation)
	{
		CompilerRef compiler = compilerNew(compilerPath.c_str());
		compilerRun(compiler, widen(readFile(filePath)), widen(filePath), widen(directoryPath), widen(libraryPath));

		if (compiler->error)
		{
			printString(compiler->error->description);
		}
		else
		{
			if (argc < 3)
			{
				// Do nothing, no specific compilation parameter
				string outputPath = removeSuffix(filePath, ".") + "ycode";
				compilerSaveResultToFile(compiler, outputPath.c_str());
			}
			else if (argv[2][0] == 'c')
			{
				// We just compiled a compiler, save its versioned bytecode somewhere
				compilerSaveResultWithVersionedFileNameToPath(compiler, directoryPath.c_str());
			}
			else if (argv[2][0] == 'r')
			{
				//char* outputPath = narrowCStringWithDifferentTail(filePath, '.', '.', "ycode");
				//compilerSaveResultToFile(compiler, outputPath);

				MachineRef machine = machineNew(compiler);
				machineRun(machine, NULL);
				delete machine;

				//free(outputPath);
			}
		}

		delete compiler;
	}
	else
	{
		MachineRef machine = machineNew(filePath.c_str());
		machineRun(machine, NULL);

		if (machine->error)
		{
			printString(machine->error->description);
		}

		delete machine;

	}

	return 0;
}