//
//  object.h
//  Yon
//
//  Created by Miso on 13.8.2009.
//  Copyright 2009 Miso Antonic. All rights reserved.
//
//

#ifndef OBJECT_H
#define OBJECT_H

#include "common.h"
//#include "assert.h"
#include "language.h"
#include <iostream>
#include "string.h"

typedef enum
{
	typeReal = 0,
    typeCall = 1,
    typeBigCall = 2,
    typeNil = 3,
    typeFunction = 4,
    typeBool = 5,
    _typeProperty = 6,
    typeNativeFunction = 7,
    typeObject = 8,
    typeString = 9,
    typeArray = 10,
    typeClass = 11,
    _typeData = 13,
    typeInternalPrototype = 14,
    numberOfObjectTypes
} ObjectType;

#define TAG_VISITED					1
#define TAG_CACHED					2
#define TAG_GARBAGE                 4
// debugging
#define TAG_DEBUG					8

struct Object;

typedef struct Slot
{
	ULONG symbol; // sizeof(ULONG) must equal sizeof(Object*) because otherwise call pooling will not work
	Object* value;
} Slot;

struct Machine;
typedef Machine* MachineRef;

typedef Object* (*NativeFunctionRef)(Object** parameters, MachineRef machine);

#define OBJECT_FIELDS \
    BYTE tag; \
    BYTE type; \
    \
    Object* parent; \
    Object* next; \
    Object* prev; \
    \
    Slot* slots; \
    ULONG slotCount; // TODO: smaller var type


typedef struct Object
{
    OBJECT_FIELDS;
} Object;
typedef Object* ObjectRef;


typedef struct RealObject
{
    OBJECT_FIELDS;
    DOUBLE value;
} RealObject;
typedef RealObject* RealObjectRef;

typedef struct ClassObject
{
    OBJECT_FIELDS;
//    Object base;
    ObjectRef instancePrototype;
} ClassObject;
typedef ClassObject* ClassObjectRef;

typedef struct StringObject
{
    OBJECT_FIELDS;
    ULONG size; 
    CHAR* data;
} StringObject;
typedef StringObject* StringObjectRef;

typedef struct ArrayObject
{
    OBJECT_FIELDS;
    ULONG size;
    ObjectRef* data;
} ArrayObject;
typedef ArrayObject* ArrayObjectRef;

struct CallObject;

typedef struct FunctionObject
{
    OBJECT_FIELDS;
    BYTE numberOfParameters; 
    BYTE* startAddress; 
    ULONG numberOfSlotsNeededForCall; 
    ULONG stackSizeNeeded; 
    ULONG* parameters; 
    CallObject* parentCall; 
    CallObject* frozenCall;
} FunctionObject;
typedef FunctionObject* FunctionObjectRef;

typedef struct NativeFunctionObject
{
    OBJECT_FIELDS;
    BYTE numberOfParameters; 
    NativeFunctionRef address; 
} NativeFunctionObject;
typedef NativeFunctionObject* NativeFunctionObjectRef;

typedef struct CallObject
{
    OBJECT_FIELDS;
    ULONG slotCapacity;
    ULONG stackCapacity;
    CallType callType;
    
    CallObject* parentCall;
    ObjectRef* stack;
    ObjectRef* stackTop; // this contains a valid value if this call is not active. If it is, stackTop variable in run function holds the valid value instead.
    CallObject* yieldedFromCall; // when using yield-from, series of calls can be frozen. If this the topmost call, yieldedFromCall contains the "from" call (point of cut of spagetti stack).
    BYTE* bytecodeAddress;
    BYTE* errorBytecode;
    
    FunctionObjectRef calledFunction;
    ObjectRef self;
    
    // This object's size is dynamic, depending on slot and stack capacity
    // Slot* slotsHere
    // ObjectRef stackHere
} CallObject;
typedef CallObject* CallObjectRef;

typedef struct PropertyObject
{
    OBJECT_FIELDS;
    ULONG readSymbol;
    ULONG writeSymbol;
} PropertyObject;
typedef PropertyObject* PropertyObjectRef;

typedef struct BoolObject
{
    OBJECT_FIELDS;
    bool value;
} BoolObject;
typedef BoolObject* BoolObjectRef;

typedef struct DataObject
{
    OBJECT_FIELDS;
    BYTE* bytes;
	ULONG size;
} DataObject;
typedef DataObject* DataObjectRef;

/*
#define addSlot(object, s, v) \
{ \
	ULONG count = ++((object)->slotCount); \
	REALLOC_ARRAY((object)->slots, Slot, count); \
	ULONG insertPoint=0; \
	while (insertPoint<count-1 && (object)->slots[insertPoint].symbol <= (s)) \
		insertPoint++; \
	\
	for (ULONG i=count-1; i>insertPoint; i--) \
		(object)->slots[i] = (object)->slots[i-1]; \
	\
	(object)->slots[insertPoint].symbol = (s); \
	(object)->slots[insertPoint].value = (v); \
}*/


#define addSlot(object, s, v) \
{ \
	ULONG count = ++((object)->slotCount); \
	REALLOC_ARRAY((object)->slots, Slot, count); \
	(object)->slots[count-1].symbol = (s); \
	(object)->slots[count-1].value = (v); \
}

static inline bool setSlot(ObjectRef object, ULONG symbol, ObjectRef value)
{
	ULONG index = object->slotCount;
	while (index--)
		if (object->slots[index].symbol == symbol)
		{
			object->slots[index].value = value;
			return true;
		}
	return false;
}

// For adding non-var and non-argument slots to context, use addSlot macro instead. since sufficient capacity is not guaranteed
//#define addSlotToCall(object, s, v) 

static inline void addSlotToCall(CallObjectRef object, ULONG s, ObjectRef v)
{
	if (setSlot((ObjectRef)object, s, v))
		return;
	ULONG count = ++object->slotCount;
//	if (count > object->slotCapacity)
//		assert(count <= object->slotCapacity);
	object->slots[count-1].symbol = s;
	object->slots[count-1].value = v;
} 


/*static inline void addSlotToCall(CallObjectRef object, ULONG s, ObjectRef v)
{
	if (setSlot((ObjectRef)object, s, v))
		return;
	ULONG count = ++object->slotCount;

	ULONG insertPoint=0; 
	while (insertPoint<count-1 && (object)->slots[insertPoint].symbol <= (s)) 
		insertPoint++; 
	
	for (ULONG i=count-1; i>insertPoint; i--)
		(object)->slots[i] = (object)->slots[i-1];
	
	(object)->slots[insertPoint].symbol = (s);
	(object)->slots[insertPoint].value = (v);
}*/

/*static inline ObjectRef getSlot(ObjectRef object, ULONG symbol)
{
	LONG min = 0;
	LONG max = object->slotCount-1;
	while (max >= min)
	{
		LONG mid = (min + max) / 2;
		if (object->slots[mid].symbol < symbol)
			min = mid + 1;
		else if (object->slots[mid].symbol > symbol)
			max = mid - 1;
		else
			return object->slots[mid].value;
	}
	return NULL;
}*/

static inline ObjectRef getSlot(ObjectRef object, ULONG symbol)
{
	ULONG index = object->slotCount;
	while (index--)
		if (object->slots[index].symbol == symbol)
			return object->slots[index].value;
	return NULL;
}

static inline LONG getSlotIndex(ObjectRef object, ULONG symbol)
{
	LONG index = object->slotCount;
	while (index--)
		if (object->slots[index].symbol == symbol)
			return index;
	return -1;
}

// Auto-optimizing
/*static inline ObjectRef getSlot(ObjectRef object, ULONG symbol)
{
	ULONG index = object->slotCount;
	while (index--)
		if (object->slots[index].symbol == symbol)
		{
			Slot result = object->slots[index];
			while (index < object->slotCount-1)
			{
				object->slots[index] = object->slots[index+1];
				index++;
			}
			object->slots[object->slotCount-1] = result;
			return result.value;
		}
	return NULL;
}*/

static inline bool setSlotRecursive(ObjectRef object, ULONG symbol, ObjectRef value)
{
	while (object)
	{
		if (setSlot(object, symbol, value))
			return true;
		object = object->parent;
	}
	return false;
}

static inline ObjectRef getSlotRecursive(ObjectRef object, ULONG symbol)
{
	while (object) // may be eventually converted to do-while loop without initial check - potencially little faster
	{
		ObjectRef result = getSlot(object, symbol);
		if (result)
			return result;
		object = object->parent;
	}
	return NULL;
}

// can only be used with call objects
static inline bool setCallSlotRecursive(CallObjectRef call, ULONG symbol, ObjectRef value)
{
	while (call)
	{
		if (setSlot((ObjectRef)call, symbol, value))
			return true;
		call = (CallObjectRef)call->parent;
	}
	return false;
}

static inline ObjectRef getCallSlotRecursive(CallObjectRef call, ULONG symbol)
{
	ObjectRef result = NULL;
	while (call)
	{
		result = getSlot((ObjectRef)call, symbol);
		if (result)
			return result;
		call = (CallObjectRef)call->parent;
	}
	return NULL;
}

#endif /* OBJECT_H */